/*
 * Copyright 1997-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

package java.awt;

import gnu.java.awt.CubicSegment;
import gnu.java.awt.LineSegment;
import gnu.java.awt.QuadSegment;
import gnu.java.awt.Segment;

import java.awt.geom.FlatteningPathIterator;
import java.awt.geom.GeneralPath;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.util.Arrays;

/**
 * The <code>BasicStroke</code> class defines a basic set of rendering
 * attributes for the outlines of graphics primitives, which are rendered
 * with a {@link Graphics2D} object that has its Stroke attribute set to 
 * this <code>BasicStroke</code>.
 * The rendering attributes defined by <code>BasicStroke</code> describe 
 * the shape of the mark made by a pen drawn along the outline of a 
 * {@link Shape} and the decorations applied at the ends and joins of 
 * path segments of the <code>Shape</code>.
 * These rendering attributes include:
 * <dl compact>
 * <dt><i>width</i>
 * <dd>The pen width, measured perpendicularly to the pen trajectory.
 * <dt><i>end caps</i>
 * <dd>The decoration applied to the ends of unclosed subpaths and
 * dash segments.  Subpaths that start and end on the same point are
 * still considered unclosed if they do not have a CLOSE segment.
 * See {@link java.awt.geom.PathIterator#SEG_CLOSE SEG_CLOSE}
 * for more information on the CLOSE segment.
 * The three different decorations are: {@link #CAP_BUTT},
 * {@link #CAP_ROUND}, and {@link #CAP_SQUARE}.
 * <dt><i>line joins</i>
 * <dd>The decoration applied at the intersection of two path segments 
 * and at the intersection of the endpoints of a subpath that is closed
 * using {@link java.awt.geom.PathIterator#SEG_CLOSE SEG_CLOSE}.
 * The three different decorations are: {@link #JOIN_BEVEL}, 
 * {@link #JOIN_MITER}, and {@link #JOIN_ROUND}.
 * <dt><i>miter limit</i>
 * <dd>The limit to trim a line join that has a JOIN_MITER decoration.
 * A line join is trimmed when the ratio of miter length to stroke
 * width is greater than the miterlimit value.  The miter length is 
 * the diagonal length of the miter, which is the distance between 
 * the inside corner and the outside corner of the intersection.
 * The smaller the angle formed by two line segments, the longer 
 * the miter length and the sharper the angle of intersection.  The
 * default miterlimit value of 10.0f causes all angles less than 
 * 11 degrees to be trimmed.  Trimming miters converts
 * the decoration of the line join to bevel.
 * <dt><i>dash attributes</i>
 * <dd>The definition of how to make a dash pattern by alternating
 * between opaque and transparent sections.
 * </dl>
 * All attributes that specify measurements and distances controlling
 * the shape of the returned outline are measured in the same
 * coordinate system as the original unstroked <code>Shape</code> 
 * argument.  When a <code>Graphics2D</code> object uses a 
 * <code>Stroke</code> object to redefine a path during the execution 
 * of one of its <code>draw</code> methods, the geometry is supplied 
 * in its original form before the <code>Graphics2D</code> transform 
 * attribute is applied.  Therefore, attributes such as the pen width 
 * are interpreted in the user space coordinate system of the 
 * <code>Graphics2D</code> object and are subject to the scaling and 
 * shearing effects of the user-space-to-device-space transform in that 
 * particular <code>Graphics2D</code>.  
 * For example, the width of a rendered shape's outline is determined
 * not only by the width attribute of this <code>BasicStroke</code>, 
 * but also by the transform attribute of the 
 * <code>Graphics2D</code> object.  Consider this code:
 * <blockquote><tt>
 *      // sets the Graphics2D object's Tranform attribute
 *	g2d.scale(10, 10);
 *      // sets the Graphics2D object's Stroke attribute
 *      g2d.setStroke(new BasicStroke(1.5f));
 * </tt></blockquote>
 * Assuming there are no other scaling transforms added to the 
 * <code>Graphics2D</code> object, the resulting line 
 * will be approximately 15 pixels wide. 
 * As the example code demonstrates, a floating-point line 
 * offers better precision, especially when large transforms are 
 * used with a <code>Graphics2D</code> object.
 * When a line is diagonal, the exact width depends on how the 
 * rendering pipeline chooses which pixels to fill as it traces the 
 * theoretical widened outline.  The choice of which pixels to turn 
 * on is affected by the antialiasing attribute because the 
 * antialiasing rendering pipeline can choose to color 
 * partially-covered pixels. 
 * <p>
 * For more information on the user space coordinate system and the 
 * rendering process, see the <code>Graphics2D</code> class comments.
 * @see Graphics2D
 * @version 1.50, 06/08/07
 * @author Jim Graham
 */
public class BasicStroke implements Stroke {

    /**
     * Joins path segments by extending their outside edges until
     * they meet.
     */
    public final static int JOIN_MITER = 0;

    /**
     * Joins path segments by rounding off the corner at a radius
     * of half the line width.
     */
    public final static int JOIN_ROUND = 1;

    /**
     * Joins path segments by connecting the outer corners of their
     * wide outlines with a straight segment.
     */
    public final static int JOIN_BEVEL = 2;

    /**
     * Ends unclosed subpaths and dash segments with no added
     * decoration.
     */
    public final static int CAP_BUTT = 0;

    /**
     * Ends unclosed subpaths and dash segments with a round
     * decoration that has a radius equal to half of the width
     * of the pen.
     */
    public final static int CAP_ROUND = 1;

    /**
     * Ends unclosed subpaths and dash segments with a square
     * projection that extends beyond the end of the segment
     * to a distance equal to half of the line width.
     */
    public final static int CAP_SQUARE = 2;

    float width;

    int join;
    int cap;
    float miterlimit;

    float dash[];
    float dash_phase;

    private Segment start, end;

    /**
     * Constructs a new <code>BasicStroke</code> with the specified
     * attributes.
     * @param width the width of this <code>BasicStroke</code>.  The
     *         width must be greater than or equal to 0.0f.  If width is
     *         set to 0.0f, the stroke is rendered as the thinnest
     *         possible line for the target device and the antialias
     *         hint setting.
     * @param cap the decoration of the ends of a <code>BasicStroke</code>
     * @param join the decoration applied where path segments meet
     * @param miterlimit the limit to trim the miter join.  The miterlimit
     *        must be greater than or equal to 1.0f.
     * @param dash the array representing the dashing pattern
     * @param dash_phase the offset to start the dashing pattern
     * @throws IllegalArgumentException if <code>width</code> is negative
     * @throws IllegalArgumentException if <code>cap</code> is not either
     *         CAP_BUTT, CAP_ROUND or CAP_SQUARE
     * @throws IllegalArgumentException if <code>miterlimit</code> is less
     *         than 1 and <code>join</code> is JOIN_MITER
     * @throws IllegalArgumentException if <code>join</code> is not
     *         either JOIN_ROUND, JOIN_BEVEL, or JOIN_MITER
     * @throws IllegalArgumentException if <code>dash_phase</code>
     *         is negative and <code>dash</code> is not <code>null</code>
     * @throws IllegalArgumentException if the length of
     *         <code>dash</code> is zero
     * @throws IllegalArgumentException if dash lengths are all zero.
     */
    public BasicStroke(float width, int cap, int join, float miterlimit,
		       float dash[], float dash_phase) {
	if (width < 0.0f) {
	    throw new IllegalArgumentException("negative width");
	}
	if (cap != CAP_BUTT && cap != CAP_ROUND && cap != CAP_SQUARE) {
	    throw new IllegalArgumentException("illegal end cap value");
	}
	if (join == JOIN_MITER) {
	    if (miterlimit < 1.0f) {
		throw new IllegalArgumentException("miter limit < 1");
	    }
	} else if (join != JOIN_ROUND && join != JOIN_BEVEL) {
	    throw new IllegalArgumentException("illegal line join value");
	}
	if (dash != null) {
	    if (dash_phase < 0.0f) {
		throw new IllegalArgumentException("negative dash phase");
	    }
	    boolean allzero = true;
	    for (int i = 0; i < dash.length; i++) {
		float d = dash[i];
		if (d > 0.0) {
		    allzero = false;
		} else if (d < 0.0) {
		    throw new IllegalArgumentException("negative dash length");
		}
	    }
	    if (allzero) {
		throw new IllegalArgumentException("dash lengths all zero");
	    }
	}
	this.width	= width;
	this.cap	= cap;
	this.join	= join;
	this.miterlimit	= miterlimit;
        if (dash != null) {
            this.dash = (float []) dash.clone();
        }
	this.dash_phase	= dash_phase;
    }

    /**
     * Constructs a solid <code>BasicStroke</code> with the specified 
     * attributes.
     * @param width the width of the <code>BasicStroke</code>
     * @param cap the decoration of the ends of a <code>BasicStroke</code>
     * @param join the decoration applied where path segments meet
     * @param miterlimit the limit to trim the miter join
     * @throws IllegalArgumentException if <code>width</code> is negative
     * @throws IllegalArgumentException if <code>cap</code> is not either
     *         CAP_BUTT, CAP_ROUND or CAP_SQUARE
     * @throws IllegalArgumentException if <code>miterlimit</code> is less
     *         than 1 and <code>join</code> is JOIN_MITER
     * @throws IllegalArgumentException if <code>join</code> is not
     *         either JOIN_ROUND, JOIN_BEVEL, or JOIN_MITER
     */
    public BasicStroke(float width, int cap, int join, float miterlimit) {
	this(width, cap, join, miterlimit, null, 0.0f);
    }

    /**
     * Constructs a solid <code>BasicStroke</code> with the specified 
     * attributes.  The <code>miterlimit</code> parameter is 
     * unnecessary in cases where the default is allowable or the 
     * line joins are not specified as JOIN_MITER.
     * @param width the width of the <code>BasicStroke</code>
     * @param cap the decoration of the ends of a <code>BasicStroke</code>
     * @param join the decoration applied where path segments meet
     * @throws IllegalArgumentException if <code>width</code> is negative
     * @throws IllegalArgumentException if <code>cap</code> is not either
     *         CAP_BUTT, CAP_ROUND or CAP_SQUARE
     * @throws IllegalArgumentException if <code>join</code> is not
     *         either JOIN_ROUND, JOIN_BEVEL, or JOIN_MITER
     */
    public BasicStroke(float width, int cap, int join) {
	this(width, cap, join, 10.0f, null, 0.0f);
    }

    /**
     * Constructs a solid <code>BasicStroke</code> with the specified 
     * line width and with default values for the cap and join 
     * styles.
     * @param width the width of the <code>BasicStroke</code>
     * @throws IllegalArgumentException if <code>width</code> is negative
     */
    public BasicStroke(float width) {
	this(width, CAP_SQUARE, JOIN_MITER, 10.0f, null, 0.0f);
    }

    /**
     * Constructs a new <code>BasicStroke</code> with defaults for all 
     * attributes.
     * The default attributes are a solid line of width 1.0, CAP_SQUARE,
     * JOIN_MITER, a miter limit of 10.0.
     */
    public BasicStroke() {
	this(1.0f, CAP_SQUARE, JOIN_MITER, 10.0f, null, 0.0f);
    }


    /**
     * Returns a <code>Shape</code> whose interior defines the 
     * stroked outline of a specified <code>Shape</code>.
     * @param s the <code>Shape</code> boundary be stroked
     * @return the <code>Shape</code> of the stroked outline.
     */
    public Shape createStrokedShape(Shape s) {
        PathIterator pi = s.getPathIterator(null);

        if( dash == null )
            return solidStroke(pi);

        return dashedStroke(pi);
    }

    /**
     * Returns the line width.  Line width is represented in user space, 
     * which is the default-coordinate space used by Java 2D.  See the
     * <code>Graphics2D</code> class comments for more information on
     * the user space coordinate system.
     * @return the line width of this <code>BasicStroke</code>.
     * @see Graphics2D
     */
    public float getLineWidth() {
	return width;
    }

    /**
     * Returns the end cap style.
     * @return the end cap style of this <code>BasicStroke</code> as one
     * of the static <code>int</code> values that define possible end cap
     * styles.
     */
    public int getEndCap() {
	return cap;
    }

    /**
     * Returns the line join style.
     * @return the line join style of the <code>BasicStroke</code> as one
     * of the static <code>int</code> values that define possible line
     * join styles.
     */
    public int getLineJoin() {
	return join;
    }

    /**
     * Returns the limit of miter joins.
     * @return the limit of miter joins of the <code>BasicStroke</code>.
     */
    public float getMiterLimit() {
	return miterlimit;
    }

    /**
     * Returns the array representing the lengths of the dash segments.
     * Alternate entries in the array represent the user space lengths
     * of the opaque and transparent segments of the dashes.
     * As the pen moves along the outline of the <code>Shape</code>
     * to be stroked, the user space
     * distance that the pen travels is accumulated.  The distance
     * value is used to index into the dash array.
     * The pen is opaque when its current cumulative distance maps
     * to an even element of the dash array and transparent otherwise.
     * @return the dash array.
     */
    public float[] getDashArray() {
        if (dash == null) {
            return null;
        }

        return (float[]) dash.clone();
    }

    /**
     * Returns the current dash phase.
     * The dash phase is a distance specified in user coordinates that 
     * represents an offset into the dashing pattern. In other words, the dash 
     * phase defines the point in the dashing pattern that will correspond to 
     * the beginning of the stroke.
     * @return the dash phase as a <code>float</code> value.
     */
    public float getDashPhase() {
	return dash_phase;
    }

    /**
     * Returns the hashcode for this stroke.
     * @return      a hash code for this stroke.
     */
    public int hashCode() {
	int hash = Float.floatToIntBits(width);
	hash = hash * 31 + join;
	hash = hash * 31 + cap;
	hash = hash * 31 + Float.floatToIntBits(miterlimit);
	if (dash != null) {
	    hash = hash * 31 + Float.floatToIntBits(dash_phase);
	    for (int i = 0; i < dash.length; i++) {
		hash = hash * 31 + Float.floatToIntBits(dash[i]);
	    }
	}
	return hash;
    }

    /**
     * Returns true if this BasicStroke represents the same
     * stroking operation as the given argument.
     */
   /**
    * Tests if a specified object is equal to this <code>BasicStroke</code>
    * by first testing if it is a <code>BasicStroke</code> and then comparing 
    * its width, join, cap, miter limit, dash, and dash phase attributes with 
    * those of this <code>BasicStroke</code>.
    * @param  obj the specified object to compare to this 
    *              <code>BasicStroke</code>
    * @return <code>true</code> if the width, join, cap, miter limit, dash, and
    *            dash phase are the same for both objects;
    *            <code>false</code> otherwise.
    */
    public boolean equals(Object obj) {
        if (!(obj instanceof BasicStroke)) {
            return false;
        }

        BasicStroke bs = (BasicStroke) obj;
        if (width != bs.width) {
            return false;
        }

        if (join != bs.join) {
            return false;
        }

        if (cap != bs.cap) {
            return false;
        }

        if (miterlimit != bs.miterlimit) {
            return false;
        }

        if (dash != null) {
	    if (dash_phase != bs.dash_phase) {
		return false;
	    }

	    if (!java.util.Arrays.equals(dash, bs.dash)) {
		return false;
	    }
        }
        else if (bs.dash != null) {
            return false;
        }

        return true;
    }

    private Shape solidStroke(PathIterator pi) {
        double[] coords = new double[6];
        double x, y, x0, y0;
        boolean pathOpen = false;
        GeneralPath output = new GeneralPath ();
        Segment[] p;
        x = x0 = y = y0 = 0;

        while (!pi.isDone()) {
            switch (pi.currentSegment(coords)) {
                case PathIterator.SEG_MOVETO:
                    x0 = x = coords[0];
                    y0 = y = coords[1];
                    if (pathOpen) {
                        capEnds();
                        convertPath(output, start);
                        start = end = null;
                        pathOpen = false;
                    }
                    break;

                case PathIterator.SEG_LINETO:
                    p = (new LineSegment(x, y, coords[0], coords[1])).
                    getDisplacedSegments(width/2.0);
                    if (!pathOpen) {
                        start = p[0];
                        end = p[1];
                        pathOpen = true;
                    } else {
                      addSegments(p);
                    }

                    x = coords[0];
                    y = coords[1];
                    break;

                case PathIterator.SEG_QUADTO:
                    p = (new QuadSegment(x, y, coords[0], coords[1], coords[2],
                                         coords[3])).getDisplacedSegments(width/2.0);
                    if (!pathOpen) {
                        start = p[0];
                        end = p[1];
                        pathOpen = true;
                    } else {
                        addSegments(p);
                    }

                    x = coords[2];
                    y = coords[3];
                    break;

                case PathIterator.SEG_CUBICTO:
                    p = new CubicSegment(x, y, coords[0], coords[1],
                                         coords[2], coords[3],
                                         coords[4], coords[5]).getDisplacedSegments(width/2.0);
                    if (!pathOpen) {
                        start = p[0];
                        end = p[1];
                        pathOpen = true;
                    } else {
                        addSegments(p);
                    }

                    x = coords[4];
                    y = coords[5];
                    break;

                case PathIterator.SEG_CLOSE:
                    if (x == x0 && y == y0) {
                        joinSegments(new Segment[] { start.first, end.first });
                    } else {
                        p = (new LineSegment(x, y, x0, y0)).getDisplacedSegments(width / 2.0);
                        addSegments(p);
                    }
                    convertPath(output, start);
                    convertPath(output, end);
                    start = end = null;
                    pathOpen = false;
                    output.setWindingRule(GeneralPath.WIND_EVEN_ODD);
                    break;
            }
            pi.next();
        }

        if (pathOpen) {
            capEnds();
            convertPath(output, start);
        }
        return output;
    }

    private Shape dashedStroke(PathIterator pi)
    {
        // The choice of (flatnessSq == width / 3) is made to be consistent with
        // the flattening in CubicSegment.getDisplacedSegments
        FlatteningPathIterator flat = new FlatteningPathIterator(pi,
                                                                 Math.sqrt(width / 3));

        // Holds the endpoint of the current segment (or piece of a segment)
        double[] coords = new double[2];

        // Holds end of the last segment
        double x, y, x0, y0;
        x = x0 = y = y0 = 0;

        // Various useful flags
        boolean pathOpen = false;
        boolean dashOn = true;
        boolean offsetting = (dash_phase != 0);

        // How far we are into the current dash
        double distance = 0;
        int dashIndex = 0;

        // And variables to hold the final output
        GeneralPath output = new GeneralPath();
        Segment[] p;

        // Iterate over the FlatteningPathIterator
        while (! flat.isDone()) {
            switch (flat.currentSegment(coords)) {
                case PathIterator.SEG_MOVETO:
                    x0 = x = coords[0];
                    y0 = y = coords[1];

                    if (pathOpen) {
                        capEnds();
                        convertPath(output, start);
                        start = end = null;
                        pathOpen = false;
                    }

                    break;

                case PathIterator.SEG_LINETO:
                    boolean segmentConsumed = false;

                    while (! segmentConsumed) {
                        // Find the total remaining length of this segment
                        double segLength = Math.sqrt((x - coords[0]) * (x - coords[0])
                                                     + (y - coords[1])
                                                     * (y - coords[1]));
                        boolean spanBoundary = true;
                        double[] segmentEnd = null;

                        // The current segment fits entirely inside the current dash
                        if ((offsetting && distance + segLength <= dash_phase)
                            || distance + segLength <= dash[dashIndex]) {
                              spanBoundary = false;

                        } else {
                            // Otherwise, we need to split the segment in two, as this
                            // segment spans a dash boundry
                            segmentEnd = (double[]) coords.clone();

                            // Calculate the remaining distance in this dash,
                            // and coordinates of the dash boundary
                            double reqLength;
                            if (offsetting)
                                reqLength = dash_phase - distance;
                            else
                                reqLength = dash[dashIndex] - distance;

                                coords[0] = x + ((coords[0] - x) * reqLength / segLength);
                                coords[1] = y + ((coords[1] - y) * reqLength / segLength);
                        }

                        if (offsetting || ! dashOn) {
                            // Dash is off, or we are in offset - treat this as a
                            // moveTo
                            x0 = x = coords[0];
                            y0 = y = coords[1];

                            if (pathOpen) {
                                capEnds();
                                convertPath(output, start);
                                start = end = null;
                                pathOpen = false;
                            }
                        } else {
                            // Dash is on - treat this as a lineTo
                            p = (new LineSegment(x, y, coords[0], coords[1])).getDisplacedSegments(width / 2.0);

                            if (! pathOpen) {
                                start = p[0];
                                end = p[1];
                                pathOpen = true;
                            } else {
                                addSegments(p);
                            }

                            x = coords[0];
                            y = coords[1];
                        }

                        // Update variables depending on whether we spanned a
                        // dash boundary or not
                        if (! spanBoundary) {
                            distance += segLength;
                            segmentConsumed = true;
                        } else {
                            if (offsetting)
                                offsetting = false;
                            dashOn = ! dashOn;
                            distance = 0;
                            coords = segmentEnd;

                            if (dashIndex + 1 == dash.length)
                                dashIndex = 0;
                            else
                                dashIndex++;

                            // Since the value of segmentConsumed is still false,
                            // the next run of the while loop will complete the segment
                        }
                    }
                    break;

            // This is a flattened path, so we don't need to deal with curves
            }
        flat.next();
        }

        if (pathOpen) {
            capEnds();
            convertPath(output, start);
        }
        return output;
     }

    /**
     * Cap the ends of the path (joining the start and end list of segments)
     */
    private void capEnds() {
        Segment returnPath = end.last;

        end.reverseAll(); // reverse the path.
        end = null;
        capEnd(start, returnPath);
        start.last = returnPath.last;
        end = null;

        capEnd(start, start);
    }

    /**
     * Append the Segments in s to the GeneralPath p
     */
    private void convertPath(GeneralPath p, Segment s) {
        Segment v = s;
        p.moveTo((float)s.P1.getX(), (float)s.P1.getY());

        do {
            if(v instanceof LineSegment)
                p.lineTo((float)v.P2.getX(), (float)v.P2.getY());
            else if(v instanceof QuadSegment)
                p.quadTo((float)((QuadSegment)v).cp.getX(),
                         (float)((QuadSegment)v).cp.getY(),
                         (float)v.P2.getX(), 
                         (float)v.P2.getY());
            else if(v instanceof CubicSegment)
                p.curveTo((float)((CubicSegment)v).cp1.getX(),
                          (float)((CubicSegment)v).cp1.getY(),
                          (float)((CubicSegment)v).cp2.getX(),
                          (float)((CubicSegment)v).cp2.getY(),
                          (float)v.P2.getX(), 
                          (float)v.P2.getY());
            v = v.next;
        } while(v != s && v != null);

        p.closePath();
    }
  
    /**
     * Add the segments to start and end (the inner and outer edges of the stroke) 
     */
    private void addSegments(Segment[] segments) {
        joinSegments(segments);
        start.add(segments[0]);
        end.add(segments[1]);
    }

    private void joinSegments(Segment[] segments) {
        double[] p0 = start.last.cp2();
        double[] p1 = new double[]{start.last.P2.getX(), start.last.P2.getY()};
        double[] p2 = new double[]{segments[0].first.P1.getX(), segments[0].first.P1.getY()};
        double[] p3 = segments[0].cp1();
        Point2D p;

        p = lineIntersection(p0[0],p0[1],p1[0],p1[1],
                             p2[0],p2[1],p3[0],p3[1], false);

        double det = (p1[0] - p0[0])*(p3[1] - p2[1]) - 
                     (p3[0] - p2[0])*(p1[1] - p0[1]);

        if (det > 0) {
            // start and segment[0] form the 'inner' part of a join, 
            // connect the overlapping segments
            joinInnerSegments(start, segments[0], p);
            joinOuterSegments(end, segments[1], p);
        } else {
            // end and segment[1] form the 'inner' part 
            joinInnerSegments(end, segments[1], p);
            joinOuterSegments(start, segments[0], p);
        }
    }

    /**
     * Make a cap between a and b segments, 
     * where a-->b is the direction of iteration.
     */
    private void capEnd(Segment a, Segment b) {
        double[] p0, p1;
        double dx, dy, l;
        Point2D c1,c2;

        switch (cap) {
            case CAP_BUTT:
                a.add(new LineSegment(a.last.P2, b.P1));
                break;

            case CAP_SQUARE:
                p0 = a.last.cp2();
                p1 = new double[]{a.last.P2.getX(), a.last.P2.getY()};
                dx = p1[0] - p0[0];
                dy = p1[1] - p0[1];
                l = Math.sqrt(dx * dx + dy * dy);
                dx = 0.5*width*dx/l;
                dy = 0.5*width*dy/l;
                c1 = new Point2D.Double(p1[0] + dx, p1[1] + dy);
                c2 = new Point2D.Double(b.P1.getX() + dx, b.P1.getY() + dy);
                a.add(new LineSegment(a.last.P2, c1));
                a.add(new LineSegment(c1, c2));
                a.add(new LineSegment(c2, b.P1));
                break;

            case CAP_ROUND:
                p0 = a.last.cp2();
                p1 = new double[]{a.last.P2.getX(), a.last.P2.getY()};
                dx = p1[0] - p0[0];
                dy = p1[1] - p0[1];
                if (dx != 0 && dy != 0) {
                    l = Math.sqrt(dx * dx + dy * dy);
                    dx = (2.0/3.0)*width*dx/l;
                    dy = (2.0/3.0)*width*dy/l;
                }
        
                c1 = new Point2D.Double(p1[0] + dx, p1[1] + dy);
                c2 = new Point2D.Double(b.P1.getX() + dx, b.P1.getY() + dy);
                a.add(new CubicSegment(a.last.P2, c1, c2, b.P1));
                break;
        }
        a.add(b);
    }

    /**
     * Returns the intersection of two lines, or null if there isn't one.
     * @param infinite - true if the lines should be regarded as infinite, false
     * if the intersection must be within the given segments.
     * @return a Point2D or null.
     */
    private Point2D lineIntersection(double X1, double Y1, 
                                     double X2, double Y2, 
                                     double X3, double Y3, 
                                     double X4, double Y4,
                                     boolean infinite) {
        double x1 = X1;
        double y1 = Y1;
        double rx = X2 - x1;
        double ry = Y2 - y1;

        double x2 = X3;
        double y2 = Y3;
        double sx = X4 - x2;
        double sy = Y4 - y2;

        double determinant = sx * ry - sy * rx;
        double nom = (sx * (y2 - y1) + sy * (x1 - x2));

        // lines can be considered parallel.
        if (Math.abs(determinant) < 1E-6)
            return null;

        nom = nom / determinant;

        // check if lines are within the bounds
        if(!infinite && (nom > 1.0 || nom < 0.0))
            return null;

        return new Point2D.Double(x1 + nom * rx, y1 + nom * ry);
    }

    /**
     * Join a and b segments, where a-->b is the direction of iteration.
     *
     * insideP is the inside intersection point of the join, needed for
     * calculating miter lengths.
     */
    private void joinOuterSegments(Segment a, Segment b, Point2D insideP) {
        double[] p0, p1;
        double dx, dy, l;
        Point2D c1,c2;

        switch (join) {
            case JOIN_MITER:
                p0 = a.last.cp2();
                p1 = new double[]{a.last.P2.getX(), a.last.P2.getY()};
                double[] p2 = new double[]{b.P1.getX(), b.P1.getY()};
                double[] p3 = b.cp1();
                Point2D p = lineIntersection(p0[0],p0[1],p1[0],p1[1],p2[0],p2[1],p3[0],p3[1], true);
                if( p == null || insideP == null )
                    a.add(new LineSegment(a.last.P2, b.P1));
                else if((p.distance(insideP)/width) < miterlimit) {
                    a.add(new LineSegment(a.last.P2, p));
                    a.add(new LineSegment(p, b.P1));
                 } else {
                     // outside miter limit, do a bevel join.
                     a.add(new LineSegment(a.last.P2, b.P1));
                 }
                 break;

            case JOIN_ROUND:
                p0 = a.last.cp2();
                p1 = new double[]{a.last.P2.getX(), a.last.P2.getY()};
                dx = p1[0] - p0[0];
                dy = p1[1] - p0[1];
                l = Math.sqrt(dx * dx + dy * dy);
                dx = 0.5*width*dx/l;
                dy = 0.5*width*dy/l;
                c1 = new Point2D.Double(p1[0] + dx, p1[1] + dy);

                p0 = new double[]{b.P1.getX(), b.P1.getY()};
                p1 = b.cp1();

                dx = p0[0] - p1[0]; // backwards direction.
                dy = p0[1] - p1[1];
                l = Math.sqrt(dx * dx + dy * dy);
                dx = 0.5*width*dx/l;
                dy = 0.5*width*dy/l;
                c2 = new Point2D.Double(p0[0] + dx, p0[1] + dy);
                a.add(new CubicSegment(a.last.P2, c1, c2, b.P1));
                break;

            case JOIN_BEVEL:
                a.add(new LineSegment(a.last.P2, b.P1));
                break;
        }
    }

    /**
     * Join a and b segments, removing any overlap
     */
    private void joinInnerSegments(Segment a, Segment b, Point2D p) {
        double[] p0 = a.last.cp2();
        double[] p1 = new double[] { a.last.P2.getX(), a.last.P2.getY() };
        double[] p2 = new double[] { b.P1.getX(), b.P1.getY() };
        double[] p3 = b.cp1();

        if (p == null) {
            // Dodgy.
            a.add(new LineSegment(a.last.P2, b.P1));
            p = new Point2D.Double((b.P1.getX() + a.last.P2.getX()) / 2.0,
                                   (b.P1.getY() + a.last.P2.getY()) / 2.0);
        } else {
            // This assumes segments a and b are single segments, which is
            // incorrect - if they are a linked list of segments (ie, passed in
            // from a flattening operation), this produces strange results!!
            a.last.P2 = b.P1 = p;
        }
    }
}
